package ru.dbotthepony.mc.otm.network

import io.netty.buffer.ByteBufAllocator
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import it.unimi.dsi.fastutil.objects.ObjectOpenHashSet
import net.minecraft.core.RegistryAccess
import net.minecraft.network.RegistryFriendlyByteBuf
import net.minecraft.world.item.ItemStack
import net.neoforged.neoforge.network.connection.ConnectionType
import ru.dbotthepony.kommons.collect.ListenableMap
import ru.dbotthepony.kommons.collect.ListenableSet
import ru.dbotthepony.kommons.util.Delegate
import ru.dbotthepony.kommons.util.DelegateGetter
import ru.dbotthepony.kommons.util.DelegateSetter
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kommons.util.Listenable
import ru.dbotthepony.kommons.util.ListenableDelegate
import ru.dbotthepony.kommons.util.Observer
import ru.dbotthepony.kommons.util.ValueObserver
import ru.dbotthepony.mc.otm.core.math.Decimal
import java.io.Closeable
import java.util.*
import java.util.concurrent.ConcurrentLinkedQueue
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.locks.ReentrantLock
import java.util.function.BooleanSupplier
import java.util.function.Consumer
import java.util.function.DoubleSupplier
import java.util.function.IntSupplier
import java.util.function.LongSupplier
import java.util.function.Supplier
import kotlin.concurrent.withLock

/**
 * Universal, one-to-many delegate/value synchronizer.
 *
 * Values and delegates can be attached using [add], or by constructing subclassed directly.
 * Delta changes are tracked by [DelegateSyncher.Remote] instances.
 *
 * In general, this class is not meant to be _structurally_ concurrently mutated by different threads,
 * to avoid structure corruption a lock is employed.
 *
 * Attached delegates can be safely mutated by multiple threads concurrently
 * (attached [ListenableDelegate] can call [Listenable.addListener] callback from different threads).
 */
@Suppress("UNCHECKED_CAST", "UNUSED")
class DelegateSyncher : Observer {
	private val lock = ReentrantLock()
	private val slots = ArrayList<AbstractSlot?>()
	private val gaps = IntAVLTreeSet()
	private val observers = ArrayList<AbstractSlot>()
	private val remotes = ArrayList<Remote>()
	private var isRemote = false

	override fun observe(): Boolean {
		var any = false
		observers.forEach { any = it.observe() || any }
		return any
	}

	fun read(stream: RegistryFriendlyByteBuf) {
		isRemote = true

		var readID = stream.readVarInt()

		while (readID > 0) {
			val slot = slots.getOrNull(readID - 1) ?: throw IndexOutOfBoundsException("Unknown networked slot ${readID - 1}!")
			slot.read(stream)
			readID = stream.readVarInt()
		}
	}

	fun read(registry: RegistryAccess, bytes: ByteArrayList) {
		return decodePayload(registry, bytes) {
			read(RegistryFriendlyByteBuf(it, registry, ConnectionType.NEOFORGE))
		}
	}

	abstract inner class AbstractSlot : Closeable, Observer, Comparable<AbstractSlot> {
		val id: Int

		protected val isRemoved = AtomicBoolean()

		fun remove() {
			if (isRemoved.compareAndSet(false, true)) {
				lock.withLock {
					slots[id] = null
					gaps.add(id)

					remoteSlots.forEach {
						it.remove()
					}

					remoteSlots.clear()
					remove0()
				}
			}
		}

		final override fun close() {
			return remove()
		}

		final override fun compareTo(other: AbstractSlot): Int {
			return id.compareTo(other.id)
		}

		protected abstract fun remove0()
		internal abstract fun read(stream: RegistryFriendlyByteBuf)
		internal abstract fun write(stream: RegistryFriendlyByteBuf)

		internal val remoteSlots = CopyOnWriteArrayList<Remote.RemoteSlot>()

		protected fun markDirty() {
			if (!isRemote && !isRemoved.get()) {
				remoteSlots.forEach {
					it.markDirty()
				}
			}
		}

		init {
			lock.withLock {
				if (gaps.isNotEmpty()) {
					val gap = gaps.firstInt()
					gaps.remove(gap)
					id = gap
					check(slots[id] == null) { "Expected slot $id to be empty" }
					slots[id] = this
				} else {
					id = slots.size
					slots.add(this)
				}

				remotes.forEach {
					it.addSlot(this)
				}
			}
		}
	}

	inner class Slot<V>(val delegate: ListenableDelegate<V>, val codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>) : AbstractSlot(), ListenableDelegate<V> {
		constructor(delegate: Supplier<V>, codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>) : this(ListenableDelegate.CustomShadow(delegate, codec::compare, codec::copy), codec)

		private val l = delegate.addListener(Consumer {
			markDirty()
			listeners.accept(it)
		})

		init {
			if (delegate is ListenableDelegate.AbstractShadow) {
				observers.add(this)
			}
		}

		private val listeners = Listenable.Impl<V>()

		override fun remove0() {
			l.remove()
			listeners.clear()
		}

		override fun read(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			delegate.accept(codec.decode(stream))
		}

		override fun write(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			codec.encode(stream, delegate.get())
		}

		override fun get(): V {
			check(!isRemoved.get()) { "This network slot was removed" }
			return delegate.get()
		}

		override fun accept(t: V) {
			check(!isRemoved.get()) { "This network slot was removed" }
			delegate.accept(t)
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return listeners.addListener(listener)
		}

		override fun observe(): Boolean {
			check(!isRemoved.get()) { "This network slot was removed" }

			if (delegate is ListenableDelegate.AbstractShadow) {
				return delegate.observe()
			}

			return false
		}
	}

	inner class ObservedSlot<V>(val delegate: Delegate<V>, val codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>) : AbstractSlot(), ListenableDelegate<V>, ValueObserver<V> {
		private val listeners = Listenable.Impl<V>()
		private var observed: Any? = Mark
		private val observeLock = ReentrantLock()

		init {
			observers.add(this)
		}

		override fun remove0() {
			listeners.clear()
		}

		override fun read(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			val read = codec.decode(stream)
			observed = codec.copy(read)
			delegate.accept(read)
		}

		override fun write(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			codec.encode(stream, delegate.get())
		}

		override fun observe(): Boolean {
			val get = delegate.get()

			if (observed === Mark || !codec.compare(get, observed as V)) {
				observeLock.withLock {
					if (observed === Mark || !codec.compare(get, observed as V)) {
						observed = codec.copy(get)
						markDirty()
						listeners.accept(get)
					}

					return true
				}
			}

			return false
		}

		override fun getAndObserve(): Pair<V, Boolean> {
			val get = delegate.get()

			if (observed === Mark || !codec.compare(get, observed as V)) {
				observeLock.withLock {
					if (observed === Mark || !codec.compare(get, observed as V)) {
						observed = codec.copy(get)
						markDirty()
						listeners.accept(get)
					}

					return get to  true
				}
			}

			return get to false
		}

		override fun get(): V {
			check(!isRemoved.get()) { "This network slot was removed" }
			return delegate.get()
		}

		override fun accept(t: V) {
			check(!isRemoved.get()) { "This network slot was removed" }
			delegate.accept(t)
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return listeners.addListener(listener)
		}
	}

	private object Mark

	internal data class SetAction<V>(val value: KOptional<V>, val action: Int)

	inner class SetSlot<V>(val delegate: ListenableSet<V>, val codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>) : AbstractSlot() {
		private val listener = object : ListenableSet.SetListener<V> {
			override fun onClear() {
				remoteSlots.forEach {
					(it as Remote.RemoteSetSlot<V>).changelist.clear()

					synchronized(it.changelistLock) {
						it.changelist.add(SetAction(KOptional(), CLEAR))
					}

					listeners.forEach {
						it.listener.onClear()
					}

					it.markDirty()
				}
			}

			override fun onValueAdded(element: V) {
				remoteSlots.forEach {
					it as Remote.RemoteSetSlot<V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.value.valueEquals(element) }
						it.changelist.add(SetAction(KOptional(codec.copy(element)), ADD))
					}

					listeners.forEach {
						it.listener.onValueAdded(element)
					}

					it.markDirty()
				}
			}

			override fun onValueRemoved(element: V) {
				remoteSlots.forEach {
					it as Remote.RemoteSetSlot<V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.value.valueEquals(element) }
						it.changelist.add(SetAction(KOptional(codec.copy(element)), REMOVE))
					}

					listeners.forEach {
						it.listener.onValueRemoved(element)
					}

					it.markDirty()
				}
			}
		}

		private val l = delegate.addListener(listener)
		private val listeners = CopyOnWriteArrayList<Listener>()

		override fun remove0() {
			l.remove()
			listeners.clear()
		}

		private inner class Listener(val listener: ListenableSet.SetListener<V>): Listenable.L {
			private var isRemoved = false

			init {
				listeners.add(this)
			}

			override fun remove() {
				if (!isRemoved) {
					isRemoved = true
					listeners.remove(this)
				}
			}
		}

		fun addListener(listener: ListenableSet.SetListener<V>): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return Listener(listener)
		}

		fun addListener(listener: Runnable): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return Listener(ListenableSet.RunnableAdapter(listener))
		}

		override fun read(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			var action = stream.readByte().toInt()

			while (true) {
				when (action) {
					ADD -> delegate.add(codec.decode(stream))
					REMOVE -> delegate.remove(codec.decode(stream))
					CLEAR -> delegate.clear()
					else -> break
				}

				action = stream.readByte().toInt()
			}
		}

		override fun write(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			throw RuntimeException("unreachable code")
		}

		override fun observe(): Boolean {
			check(!isRemoved.get()) { "This network slot was removed" }
			return false
		}
	}

	internal data class MapAction<K, V>(val key: KOptional<K>, val value: KOptional<V>, val action: Int)

	inner class MapSlot<K, V>(val delegate: ListenableMap<K, V>, val keyCodec: MatteryStreamCodec<in RegistryFriendlyByteBuf, K>, val valueCodec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>) : AbstractSlot() {
		private val listener = object : ListenableMap.MapListener<K, V> {
			override fun onClear() {
				remoteSlots.forEach {
					it as Remote.RemoteMapSlot<K, V>

					synchronized(it.changelistLock) {
						it.changelist.clear()
						it.changelist.add(MapAction(KOptional(), KOptional(), CLEAR))
					}

					listeners.forEach {
						it.listener.onClear()
					}

					it.markDirty()
				}
			}

			override fun onValueAdded(key: K, value: V) {
				remoteSlots.forEach {
					it as Remote.RemoteMapSlot<K, V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.key.valueEquals(key) }
						it.changelist.add(MapAction(KOptional(keyCodec.copy(key)), KOptional(valueCodec.copy(value)), ADD))
					}

					listeners.forEach {
						it.listener.onValueAdded(key, value)
					}

					it.markDirty()
				}
			}

			override fun onValueRemoved(key: K, value: V) {
				remoteSlots.forEach {
					it as Remote.RemoteMapSlot<K, V>

					synchronized(it.changelistLock) {
						it.changelist.removeIf { it.key.valueEquals(key) }
						it.changelist.add(MapAction(KOptional(keyCodec.copy(key)), KOptional(), REMOVE))
					}

					listeners.forEach {
						it.listener.onValueRemoved(key, value)
					}

					it.markDirty()
				}
			}
		}

		private val l = delegate.addListener(listener)
		private val listeners = CopyOnWriteArrayList<Listener>()

		override fun remove0() {
			l.remove()
			listeners.clear()
		}

		private inner class Listener(val listener: ListenableMap.MapListener<K, V>): Listenable.L {
			private var isRemoved = false

			init {
				listeners.add(this)
			}

			override fun remove() {
				if (!isRemoved) {
					isRemoved = true
					listeners.remove(this)
				}
			}
		}

		fun addListener(listener: ListenableMap.MapListener<K, V>): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return Listener(listener)
		}

		fun addListener(listener: Runnable): Listenable.L {
			check(!isRemoved.get()) { "This network slot was removed" }
			return Listener(ListenableMap.RunnableAdapter(listener))
		}

		override fun read(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			var action = stream.readByte().toInt()

			while (true) {
				when (action) {
					ADD -> delegate.put(keyCodec.decode(stream), valueCodec.decode(stream))
					REMOVE -> delegate.remove(keyCodec.decode(stream))
					CLEAR -> delegate.clear()
					else -> break
				}

				action = stream.readByte().toInt()
			}
		}

		override fun write(stream: RegistryFriendlyByteBuf) {
			check(!isRemoved.get()) { "This network slot was removed" }
			throw RuntimeException("unreachable code")
		}

		override fun observe(): Boolean {
			check(!isRemoved.get()) { "This network slot was removed" }
			return false
		}
	}

	fun <V> add(value: V, codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>, setter: DelegateSetter<V> = DelegateSetter.passthrough(), getter: DelegateGetter<V> = DelegateGetter.passthrough()): Slot<V> {
		return Slot(ListenableDelegate.maskSmart(value, getter, setter), codec)
	}

	fun <V> add(delegate: ListenableDelegate<V>, codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>): Slot<V> {
		return Slot(delegate, codec)
	}

	fun <V> add(delegate: Delegate<V>, codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>): ObservedSlot<V> {
		return ObservedSlot(delegate, codec)
	}

	fun <V> add(delegate: Supplier<V>, codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>): Slot<V> {
		return computed(delegate, codec)
	}

	fun <E> add(delegate: ListenableSet<E>, codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, E>): SetSlot<E> {
		return SetSlot(delegate, codec)
	}

	fun <K, V> add(delegate: ListenableMap<K, V>, keyCodec: MatteryStreamCodec<in RegistryFriendlyByteBuf, K>, valueCodec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>): MapSlot<K, V> {
		return MapSlot(delegate, keyCodec, valueCodec)
	}

	fun computedByte(delegate: Supplier<Byte>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.BYTE)
	fun computedShort(delegate: Supplier<Short>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.SHORT)
	fun computedInt(delegate: Supplier<Int>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.INT)
	fun computedLong(delegate: Supplier<Long>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.VAR_LONG)
	fun computedFloat(delegate: Supplier<Float>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.FLOAT)
	fun computedDouble(delegate: Supplier<Double>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.DOUBLE)
	fun computedBoolean(delegate: Supplier<Boolean>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.BOOLEAN)
	fun computedString(delegate: Supplier<String>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.STRING)
	fun computedUUID(delegate: Supplier<UUID>) = Slot(ListenableDelegate.Shadow(delegate), StreamCodecs.UUID)
	fun <V> computed(delegate: Supplier<V>, codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>) = Slot(ListenableDelegate.CustomShadow(delegate, codec::compare, codec::copy), codec)

	fun computedInt(delegate: IntSupplier) = Slot(ListenableDelegate.Shadow(delegate::getAsInt), StreamCodecs.VAR_INT)
	fun computedLong(delegate: LongSupplier) = Slot(ListenableDelegate.Shadow(delegate::getAsLong), StreamCodecs.VAR_LONG)
	fun computedDouble(delegate: DoubleSupplier) = Slot(ListenableDelegate.Shadow(delegate::getAsDouble), StreamCodecs.DOUBLE)
	fun computedBoolean(delegate: BooleanSupplier) = Slot(ListenableDelegate.Shadow(delegate::getAsBoolean), StreamCodecs.BOOLEAN)

	@JvmName("vbyte")
	@JvmOverloads
	fun byte(value: Byte = 0, setter: DelegateSetter<Byte> = DelegateSetter.passthrough(), getter: DelegateGetter<Byte> = DelegateGetter.passthrough()): Slot<Byte> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.BYTE)
	}

	@JvmName("vshort")
	@JvmOverloads
	fun short(value: Short = 0, setter: DelegateSetter<Short> = DelegateSetter.passthrough(), getter: DelegateGetter<Short> = DelegateGetter.passthrough()): Slot<Short> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.SHORT)
	}

	@JvmName("vint")
	@JvmOverloads
	fun int(value: Int = 0, setter: DelegateSetter<Int> = DelegateSetter.passthrough(), getter: DelegateGetter<Int> = DelegateGetter.passthrough()): Slot<Int> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.VAR_INT)
	}

	@JvmName("vlong")
	@JvmOverloads
	fun long(value: Long = 0L, setter: DelegateSetter<Long> = DelegateSetter.passthrough(), getter: DelegateGetter<Long> = DelegateGetter.passthrough()): Slot<Long> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.VAR_LONG)
	}

	@JvmName("vfloat")
	@JvmOverloads
	fun float(value: Float = 0f, setter: DelegateSetter<Float> = DelegateSetter.passthrough(), getter: DelegateGetter<Float> = DelegateGetter.passthrough()): Slot<Float> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.FLOAT)
	}

	@JvmName("vdouble")
	@JvmOverloads
	fun double(value: Double = 0.0, setter: DelegateSetter<Double> = DelegateSetter.passthrough(), getter: DelegateGetter<Double> = DelegateGetter.passthrough()): Slot<Double> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.DOUBLE)
	}

	@JvmName("vboolean")
	@JvmOverloads
	fun boolean(value: Boolean = false, setter: DelegateSetter<Boolean> = DelegateSetter.passthrough(), getter: DelegateGetter<Boolean> = DelegateGetter.passthrough()): Slot<Boolean> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.BOOLEAN)
	}

	@JvmOverloads
	fun string(value: String, setter: DelegateSetter<String> = DelegateSetter.passthrough(), getter: DelegateGetter<String> = DelegateGetter.passthrough()): Slot<String> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.STRING)
	}

	@JvmOverloads
	fun uuid(value: UUID, setter: DelegateSetter<UUID> = DelegateSetter.passthrough(), getter: DelegateGetter<UUID> = DelegateGetter.passthrough()): Slot<UUID> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.UUID)
	}

	@JvmOverloads
	fun <E> set(codec: MatteryStreamCodec<in RegistryFriendlyByteBuf, E>, backing: MutableSet<E> = ObjectOpenHashSet()): SetSlot<E> {
		return add(ListenableSet(backing), codec)
	}

	@JvmOverloads
	fun <K, V> map(keyCodec: MatteryStreamCodec<in RegistryFriendlyByteBuf, K>, valueCodec: MatteryStreamCodec<in RegistryFriendlyByteBuf, V>, backing: MutableMap<K, V> = Object2ObjectOpenHashMap()): MapSlot<K, V> {
		return add(ListenableMap(backing), keyCodec, valueCodec)
	}

	@JvmOverloads
	fun <E : Enum<E>> enum(value: E, setter: DelegateSetter<E> = DelegateSetter.passthrough(), getter: DelegateGetter<E> = DelegateGetter.passthrough()): Slot<E> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), MatteryStreamCodec.Enum(value::class.java))
	}

	fun decimal(value: Decimal = Decimal.ZERO, setter: DelegateSetter<Decimal> = DelegateSetter.passthrough(), getter: DelegateGetter<Decimal> = DelegateGetter.passthrough()): DelegateSyncher.Slot<Decimal> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), StreamCodecs.DECIMAL)
	}

	fun computedDecimal(delegate: Supplier<Decimal>): DelegateSyncher.Slot<Decimal> {
		return add(delegate, StreamCodecs.DECIMAL)
	}

	fun item(value: ItemStack = ItemStack.EMPTY, setter: DelegateSetter<ItemStack> = DelegateSetter.passthrough(), getter: DelegateGetter<ItemStack> = DelegateGetter.passthrough()): DelegateSyncher.Slot<ItemStack> {
		return add(ListenableDelegate.maskSmart(value, getter, setter), ItemStack.STREAM_CODEC.wrap())
	}

	fun computedItem(delegate: Supplier<ItemStack>): DelegateSyncher.Slot<ItemStack> {
		return computed(delegate, ItemStack.STREAM_CODEC.wrap())
	}

	fun observedItem(value: ItemStack = ItemStack.EMPTY, setter: DelegateSetter<ItemStack> = DelegateSetter.passthrough(), getter: DelegateGetter<ItemStack> = DelegateGetter.passthrough()): DelegateSyncher.ObservedSlot<ItemStack> {
		return add(Delegate.maskSmart(value, getter, setter), ItemStack.STREAM_CODEC.wrap())
	}

	/**
	 * Remotes must be [close]d when no longer in use, otherwise they will
	 * leak memory and decrease performance of syncher.
	 *
	 * To get changes to be networked to remote client, [write] should be called.
	 */
	inner class Remote : Closeable {
		@Volatile
		private var isRemoved = false
		internal val dirty = ConcurrentLinkedQueue<RemoteSlot>()
		private val remoteSlots = CopyOnWriteArrayList<RemoteSlot>()

		internal open inner class RemoteSlot(val parent: AbstractSlot) : Comparable<RemoteSlot> {
			private val isDirty = AtomicBoolean()

			final override fun compareTo(other: RemoteSlot): Int {
				return parent.compareTo(other.parent)
			}

			init {
				remoteSlots.add(this)
				dirty.add(this)
			}

			fun markDirty() {
				if (!isRemoved && isDirty.compareAndSet(false, true)) {
					dirty.add(this)
				}
			}

			fun markClean() {
				isDirty.set(false)
			}

			open fun invalidate() {
				markDirty()
			}

			open fun remove() {
				isDirty.set(true)
				remoteSlots.remove(this)
				dirty.remove(this)
			}

			open fun write(stream: RegistryFriendlyByteBuf) {
				parent.write(stream)
			}

			override fun hashCode(): Int {
				return parent.hashCode()
			}

			override fun equals(other: Any?): Boolean {
				return this === other || other is Remote.RemoteSlot && other.parent == parent
			}
		}

		internal inner class RemoteSetSlot<V>(parent: SetSlot<V>) : RemoteSlot(parent) {
			val changelist = ArrayList<SetAction<V>>()
			val changelistLock = Any()

			override fun remove() {
				super.remove()
				changelist.clear()
			}

			override fun invalidate() {
				synchronized(changelistLock) {
					changelist.clear()
					changelist.add(SetAction(KOptional(), CLEAR))

					parent as SetSlot<V>

					for (v in parent.delegate) {
						changelist.add(SetAction(KOptional(parent.codec.copy(v)), ADD))
					}
				}

				super.invalidate()
			}

			override fun write(stream: RegistryFriendlyByteBuf) {
				synchronized(changelistLock) {
					for (change in changelist) {
						stream.writeByte(change.action)
						change.value.ifPresent { (parent as SetSlot<V>).codec.encode(stream, it) }
					}

					changelist.clear()
				}

				stream.writeByte(0)
			}
		}

		internal inner class RemoteMapSlot<K, V>(parent: MapSlot<K, V>) : RemoteSlot(parent) {
			val changelist = ArrayList<MapAction<K, V>>()
			val changelistLock = Any()

			override fun remove() {
				super.remove()
				changelist.clear()
			}

			override fun invalidate() {
				synchronized(changelistLock) {
					changelist.clear()
					changelist.add(MapAction(KOptional(), KOptional(), CLEAR))

					parent as MapSlot<K, V>

					for ((k, v) in parent.delegate) {
						changelist.add(MapAction(KOptional(parent.keyCodec.copy(k)), KOptional(parent.valueCodec.copy(v)), ADD))
					}
				}

				super.invalidate()
			}

			override fun write(stream: RegistryFriendlyByteBuf) {
				synchronized(changelistLock) {
					for (change in changelist) {
						stream.writeByte(change.action)
						change.key.ifPresent { (parent as MapSlot<K, V>).keyCodec.encode(stream, it) }
						change.value.ifPresent { (parent as MapSlot<K, V>).valueCodec.encode(stream, it) }
					}

					changelist.clear()
				}

				stream.writeByte(0)
			}
		}

		init {
			lock.withLock {
				slots.forEach {
					if (it != null) {
						addSlot(it)
					}
				}

				remotes.add(this)
			}
		}

		internal fun addSlot(slot: AbstractSlot) {
			if (slot is SetSlot<*>) {
				slot.remoteSlots.add(RemoteSetSlot(slot))
			} else if (slot is MapSlot<*, *>) {
				slot.remoteSlots.add(RemoteMapSlot(slot))
			} else {
				slot.remoteSlots.add(RemoteSlot(slot))
			}
		}

		/**
		 * Returns null if this remote is clean.
		 *
		 * [DelegateSyncher.observe] is not called automatically for performance
		 * reasons, you must call it manually.
		 */
		fun write(registry: RegistryAccess): ByteArrayList? {
			if (dirty.isNotEmpty()) {
				val sorted = ObjectAVLTreeSet<RemoteSlot>()
				var next = dirty.poll()

				while (next != null) {
					sorted.add(next)
					next.markClean()
					next = dirty.poll()
				}

				return encodePayload(registry) {
					for (slot in sorted) {
						it.writeVarInt(slot.parent.id + 1)
						slot.write(it)
					}

					it.writeByte(0)
				}
			}

			return null
		}

		/**
		 * Marks all networked slots dirty
		 */
		fun invalidate() {
			remoteSlots.forEach { it.invalidate() }
		}

		override fun close() {
			if (!isRemoved) {
				lock.withLock {
					if (!isRemoved) {
						remoteSlots.forEach {
							it.remove()
							it.parent.remoteSlots.remove(it)
						}

						remotes.remove(this)
					}
				}

				dirty.clear()
			}
		}
	}

	companion object {
		private const val END = 0
		private const val CLEAR = 1
		private const val ADD = 2
		private const val REMOVE = 3
	}
}
